@@ -1,55 +1,56 @@
-import json
 import copy
-import functools
-from logging import error, info, debug
+import json
 import re
+from logging import error, info
+
 import requests
-from .schema import Schema, Field, schema_doc, MissingField
+
 from . import DEFAULT_ACCESS_MODE
 from .re import UUID_MATCH
+from .schema import Field, MissingField, Schema, schema_doc
 
 
 class NotImplemented(Exception):
     def __init__(self, *args, **kwargs):
         Exception.__init__(self, *args, **kwargs)
 
-class InvalidData(Exception):
-    def __init__(self, *args, **kwargs):
-        Exception.__init__(self, *args, **kwargs)
 
 class ArgumentError(Exception):
     def __init__(self, *args, **kwargs):
         Exception.__init__(self, *args, **kwargs)
 
+
 class ResponseError(Exception):
     def __init__(self, *args, **kwargs):
         Exception.__init__(self, *args, **kwargs)
 
+
 class ValidationError(Exception):
     def __init__(self, *args, **kwargs):
         Exception.__init__(self, *args, **kwargs)
 
 
 ERROR_HANDLER = {
-        # Mapping of message templates provided in 412 errors from backend to
-        # Exceptions that will be raised
-        "object.not.valid": lambda msg: ValidationError(
-            "{message} ({field}={parameter})".format(**msg)),
-        "organization.not.exist": lambda msg: ValidationError(
-            "{message} ({field}={parameter})".format(**msg))
+    # Mapping of message templates provided in 412 errors from backend to
+    # Exceptions that will be raised
+    "object.not.valid": lambda msg: ValidationError(
+        "{message} ({field}={parameter})".format(**msg)
+    ),
+    "organization.not.exist": lambda msg: ValidationError(
+        "{message} ({field}={parameter})".format(**msg)
+    ),
 }
 
 
-def request(method, user_id, url, requests_common_kwargs = None, **kwargs):
+def request(method, user_id, url, requests_common_kwargs=None, **kwargs):
     """Perform requests towards API
 
-Args:
-    method (str):         POST|GET
-    user_id (int):        Act user ID
-    url (str):            Absolute URL for the endpoint
-    **kwargs (keywords):  Additional options passed to requests json parameter
-                          the following fields:
-"""
+    Args:
+        method (str):         POST|GET
+        user_id (int):        Act user ID
+        url (str):            Absolute URL for the endpoint
+        **kwargs (keywords):  Additional options passed to requests json parameter
+                              the following fields:"""
 
     if not requests_common_kwargs:
         requests_common_kwargs = {}
@@ -68,11 +69,7 @@ Args:
         requests_kwargs["headers"]["ACT-User-ID"] = str(user_id)
 
     try:
-        res = requests.request(
-            method,
-            url,
-            **requests_kwargs
-        )
+        res = requests.request(method, url, **requests_kwargs)
     except requests.exceptions.ConnectionError as e:
         raise ResponseError("Connection error {}".format(e))
 
@@ -89,22 +86,22 @@ Args:
                 raise ERROR_HANDLER[msg_template](msg)
 
         # All other, unhandled errors - log to error() and raise generic exception
-        error("Request failed: {}, {}, {}".format(
-            url, kwargs, res.status_code))
+        error("Request failed: {}, {}, {}".format(url, kwargs, res.status_code))
         raise ResponseError(res.text)
 
     elif res.status_code not in (200, 201):
-        error("Request failed: {}, {}, {}".format(
-            url, kwargs, res.status_code))
+        error("Request failed: {}, {}, {}".format(url, kwargs, res.status_code))
         raise ResponseError(
-            "Unknown response error {}: {}".format(res.status_code, res.text))
+            "Unknown response error {}: {}".format(res.status_code, res.text)
+        )
 
     try:
         return res.json()
 
     except json.decoder.JSONDecodeError:
         raise ResponseError(
-            "Error decoding response {}: {}".format(res.status_code, res.text))
+            "Error decoding response {}: {}".format(res.status_code, res.text)
+        )
 
 
 class ActResultSet(object):
@@ -112,22 +109,19 @@ class ActResultSet(object):
 
     def __init__(self, response, deserializer):
         """Initialize result set
-Args:
-    response (str):       JSON response from Act. This should include
-                          the following fields:
-                            - count: the number of entries fetched
-                            - limit: the limit in the query
-                            - responseCode: responseCode from the API
-                            - size: the total number of entries in the platform
-                            - data (array): array of entries
+        Args:
+            response (str):       JSON response from Act. This should include
+                                  the following fields:
+                                    - count: the number of entries fetched
+                                    - limit: the limit in the query
+                                    - responseCode: responseCode from the API
+                                    - size: the total number of entries in the platform
+                                    - data (array): array of entries
 
-    deserializer (class): Deseralizer class
-"""
+            deserializer (class): Deseralizer class"""
 
         if not isinstance(response["data"], list):
-            raise ResponseError(
-                "Response should be list: {}".format(
-                    response["data"]))
+            raise ResponseError("Response should be list: {}".format(response["data"]))
 
         self.data = [deserializer(**d) for d in response["data"]]
 
@@ -144,8 +138,7 @@ Args:
     def __call__(self, func, *args, **kwargs):
         """Call function on each data entry"""
 
-        self.data = [getattr(item, func)(*args, **kwargs)
-                     for item in self.data]
+        self.data = [getattr(item, func)(*args, **kwargs) for item in self.data]
         return self
 
     def __len__(self):
@@ -161,7 +154,7 @@ Args:
         return "\n".join(["{}".format(item) for item in self.data])
 
     def __bool__(self):
-        """ Return True for non empty result sets """
+        """Return True for non empty result sets"""
         if self.size > 0:
             return True
 
@@ -181,19 +174,21 @@ Args:
 
 class Config(object):
     """Config object"""
+
     act_baseurl = None
     user_id = None
     requests_common_kwargs = {}
 
     def __init__(
-            self,
-            act_baseurl,
-            user_id,
-            requests_common_kwargs = None,
-            origin_name=None,
-            origin_id=None,
-            access_mode=DEFAULT_ACCESS_MODE,
-            organization=None):
+        self,
+        act_baseurl,
+        user_id,
+        requests_common_kwargs=None,
+        origin_name=None,
+        origin_id=None,
+        access_mode=DEFAULT_ACCESS_MODE,
+        organization=None,
+    ):
         """
         act_baseurl - url to ACT instance
         use_id - ACT user ID
@@ -242,7 +237,7 @@ class ActBase(Schema):
             self.config.user_id,
             "{}/{}".format(self.config.act_baseurl, uri),
             self.config.requests_common_kwargs,
-            **kwargs
+            **kwargs,
         )
 
         return response
@@ -259,45 +254,30 @@ class ActBase(Schema):
 
     def api_delete(self, uri, params=None):
         """Send DELETE request to API
-Args:
-    uri (str):     URI (relative to base url). E.g. "v1/factType"
-    params (Dict): Parameters that are URL enncoded and sent to the API
-"""
+        Args:
+            uri (str):     URI (relative to base url). E.g. "v1/factType"
+            params (Dict): Parameters that are URL enncoded and sent to the API"""
 
         return self.api_request("DELETE", uri, params=params)
 
     def api_get(self, uri, params=None):
         """Send GET request to API
-Args:
-    uri (str):     URI (relative to base url). E.g. "v1/factType"
-    params (Dict): Parameters that are URL enncoded and sent to the API
-"""
+        Args:
+            uri (str):     URI (relative to base url). E.g. "v1/factType"
+            params (Dict): Parameters that are URL enncoded and sent to the API"""
 
         return self.api_request("GET", uri, params=params)
 
     def __eq__(self, other):
-        if not isinstance(other, self.__class__):
-            return False # Different types -> not equal
+        "Equality operator"
 
-        for field, value in self.data.items():
-            # Only compare serialized fields. The other fields
-            # have different representation if they are created locally
-            # and not recieved from the back end
-            if self.get_field(field).serializer is False:
-                continue
-            if field == "id":
-                # Facts/objects may not have an ID, unless they are returned from the backend
-                # We will check for inconsistencies below
-                continue
-            if other.data.get(field) != value:
-                return False # Different field value
+        return hash(self) == hash(other)
 
-        # Two objects where all fields are equal do not have the same id
-        if self.id and other.id and self.id != other.id:
-            raise InvalidData("Two objects with equal fields do not have the same id")
-
-        # All field values are equal
-        return True
+    def __hash__(self):
+        "__hash__ should be implemented on all derived classes"
+        raise NotImplementedError(
+            f"{self.__class__.__name__} is missing __hash__ method"
+        )
 
 
 class NameSpace(ActBase):
@@ -308,6 +288,14 @@ class NameSpace(ActBase):
         Field("id"),
     ]
 
+    def __hash__(self):
+        return hash(
+            (
+                self.__class__.__name__,
+                self.name,
+            )
+        )
+
 
 class Organization(ActBase):
     """Manage Organization"""
@@ -317,6 +305,14 @@ class Organization(ActBase):
         Field("id"),
     ]
 
+    def __hash__(self):
+        return hash(
+            (
+                self.__class__.__name__,
+                self.name,
+            )
+        )
+
     def serialize(self):
         # Return None for empty objects (non initialized objects)
         # otherwize return id or name
@@ -324,11 +320,11 @@ class Organization(ActBase):
 
 
 def origin_serializer(origin):
-        # Return None for empty objects (non initialized origins)
-        # otherwize return id or name
-        if not origin:
-            return None
-        return origin.id or origin.name
+    # Return None for empty objects (non initialized origins)
+    # otherwize return id or name
+    if not origin:
+        return None
+    return origin.id or origin.name
 
 
 class Origin(ActBase):
@@ -345,6 +341,11 @@ class Origin(ActBase):
         Field("flags", serializer=False),
     ]
 
+    def __hash__(self):
+        return hash(
+            (self.__class__.__name__, self.name, self.namespace, self.organization)
+        )
+
     @schema_doc(SCHEMA)
     def __init__(self, *args, **kwargs):
         super(Origin, self).__init__(*args, **kwargs)
@@ -353,8 +354,7 @@ class Origin(ActBase):
         """Get Origin"""
 
         if not self.id:
-            raise MissingField(
-                "Must have fact ID to get origin")
+            raise MissingField("Must have fact ID to get origin")
 
         origin = self.api_get("v1/origin/uuid/{}".format(self.id))["data"]
         self.data = {}
@@ -379,8 +379,7 @@ class Origin(ActBase):
         """Delete Origin"""
 
         if not self.id:
-            raise MissingField(
-                "Must have fact ID to delete origin")
+            raise MissingField("Must have fact ID to delete origin")
 
         origin = self.api_delete("v1/origin/uuid/{}".format(self.id))["data"]
         self.data = {}
@@ -400,3 +399,12 @@ class Comment(ActBase):
         Field("reply_to"),
         Field("origin", deserializer=Origin, serializer=False),
     ]
+
+    def __hash__(self):
+        return hash(
+            (
+                self.__class__.__name__,
+                self.comment,
+                self.timestamp,
+            )
+        )
